

class Mutation(object):
    """

    .. attribute:: old

        Used either in a deletion or replacement. When used in a replacement, this must
        be a :class:`Feature`; when used in a deletion this can be any *insertable*.

        This may be a :class:`Plasmid` to signify a removed non-integrated plasmid. (Plasmid vectors that have been
        integrated cannot be removed directly --- their contents must be removed instead.)

    .. attribute:: new

        Used either in an insertion, or replacement.

    .. attribute:: markers

        A :class:`FeatureSet`, acting as a set of selection markers. The selection markers are generally inserted as part
        of the mutation.

    .. attribute:: multiple

        A :class:`bool`, specifying whether :attr:`old` is a multiple integration site.

        This attribute is only used with a replacement.

    """

    def __init__(self, old, new, markers=None, multiple=False):
        if isinstance(old, (list, tuple)):
            old = FeatureTree(*old)
        elif old and not isinstance(old, Plasmid):
            old = FeatureTree(old)
        if isinstance(new, (list, tuple)):
            new = FeatureTree(*new)
        elif new and not isinstance(new, Plasmid):
            new = FeatureTree(new)
        if markers is not None:
            markers = FeatureTree(*markers)
        self.old = old
        self.new = new
        self.markers = markers
        self.multiple = multiple

    def __eq__(self, other):
        return isinstance(other, Mutation) and \
            self.old == other.old and \
            self.new == other.new and \
            self.markers == other.markers and \
            self.multiple == other.multiple

    def __hash__(self):
        return hash(self.old) + \
               hash(self.new) + \
               hash(self.markers) + \
               hash(self.multiple)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__,
                               ', '.join('{}={}'.format(key, repr(value))
                                         for key, value in self.__dict__.items() if value is not None))


def Ins(insert, **kwargs):
    return Mutation(None, insert, **kwargs)


def Sub(before, after, **kwargs):
    return Mutation(before, after, **kwargs)


def Del(delete, **kwargs):
    return Mutation(delete, None, **kwargs)


class MatchableMixin(object):
    def match(self, other):
        return self == other


class FeatureTree(object):
    """
    A grouping or collection of features and fusions.
    """
    def __init__(self, *contents):
        self.contents = contents

    def features(self):
        for item in self.contents:
            if isinstance(item, FeatureTree):
                # yield from item.features()
                for feature in item.features():
                    yield feature
            else:
                yield item

    def __len__(self):
        return len(self.contents)

    def __getitem__(self, item):
        return self.contents[item]

    def __iter__(self):
        return iter(self.contents)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.contents == other.contents

    def __hash__(self):
        return hash(self.contents)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, ', '.join(map(repr, self.contents)))


class FeatureSet(FeatureTree, MatchableMixin):

    def match(self, other):
        if not isinstance(other, FeatureSet):
            return False

        # feature sets must have the same number of features to match
        if len(self) != len(other):
            return False

        return all(a.match(b) for a, b in zip(self.contents, other.contents))


class Fusion(FeatureTree, MatchableMixin):
    """
    A fusion must contain at least two features.
    """

    def __hash__(self):
        return hash(self.contents)

    def contains(self, other):
        if isinstance(other, Fusion):
            for i, feature in enumerate(self.contents):
                if feature == other[0]:
                    if self.contents[i:i + len(other)] == other.contents:
                        return True
            return False
        else:
            return other in self.contents

    def match(self, other):
        if not isinstance(other, Fusion):
            return False

        # fusions must have the same number of features to match
        if len(self) != len(other):
            return False

        return all(a.match(b) for a, b in zip(self.contents, other.contents))

    def __eq__(self, other):
        return isinstance(other, Fusion) and self.contents == other.contents


class Plasmid(FeatureTree, MatchableMixin):
    def __init__(self, name, contents, site=None, markers=None):
        super(Plasmid, self).__init__(*contents if contents else ())
        if markers is not None:
            markers = FeatureSet(*markers)
        self.name = name
        self.site = site
        self.markers = markers

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return isinstance(other, Plasmid) and self.name == other.name

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, repr(self.__dict__))

    def match(self, other, match_contents=False):
        if not self == other:
            return False

        if match_contents and not super(Plasmid, self).match(other):
            return False

        return True


class Feature(MatchableMixin):
    def __init__(self, name=None, type=None, accession=None, organism=None, variant=None, range=None):
        self.name = name
        self.type = type
        self.accession = accession
        self.organism = organism
        self.variant = variant
        self.range = range

    @classmethod
    def parse(cls, string, *args, **kwargs):
        """

        ::

            >>> Feature.parse('MYO1#SGD:S000001065')
            Feature(name='MYO1', accession=Accession(database='SGD', identifier='S000001065'}))

        :param string:
        :param args:
        :param kwargs:
        :return:
        """
        from gnomic.grammar import GnomicParser
        from gnomic.semantics import DefaultSemantics

        parser = GnomicParser()
        semantics = DefaultSemantics(*args, **kwargs)
        return parser.parse(string,
                            whitespace='',
                            semantics=semantics,
                            rule_name='FEATURE')

    def match(self, other, match_variant=True):
        if not isinstance(other, Feature):
            return False

        if self.accession and other.accession:
            return self.accession == other.accession
        elif self.name:
            # only match features with the same name
            if self.name != other.name:
                return False

            # if an organism is specified, match only features with the same organism
            if self.organism and self.organism != other.organism:
                return False

            # TODO ranges
            if self.range and self.range != other.range:
                return False

            # if this feature has no variant, match any other feature; otherwise, match only features with the same
            # variant
            if not self.variant or match_variant is False:
                return True
            if self.variant == other.variant:
                return True
            return False
        else:
            # not enough information for any match
            return False

    def __eq__(self, other):
        if not isinstance(other, Feature):
            return False

        if self.accession and other.accession:
            return self.accession == other.accession and self.range == other.range
        elif self.name:
            return self.name == other.name and \
                   self.type == other.type and \
                   self.organism == other.organism and \
                   self.variant == other.variant and \
                   self.range == other.range
        # TODO range
        return False

    def __hash__(self):
        return hash(self.name) + \
               hash(self.type) + \
               hash(self.accession) + \
               hash(self.organism) + \
               hash(self.variant) + \
               hash(self.range)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__,
                               ', '.join('{}={}'.format(key, repr(value))
                                         for key, value in self.__dict__.items() if value))


class Range(object):
    """
    An inclusive range at a coding (DNA), RNA or protein level.
    """
    def __init__(self, level, start, end):
        self.level = level
        self.start = start
        self.end = end

    def __hash__(self):
        return hash(self.level) + \
               hash(self.start) + \
               hash(self.end)

    def __len__(self):
        return self.end - self.start + 1

    def __eq__(self, other):
        return isinstance(other, Range) \
               and self.level == other.level \
               and self.start == other.start\
               and self.end == other.end

    def __repr__(self):
        if self.start == self.end:
            return '{}({}, {})'.format(self.__class__.__name__, repr(self.level), self.start)
        return '{}({}, {}, {})'.format(self.__class__.__name__, repr(self.level), self.start, self.end)


class Organism(object):
    def __init__(self, name, aliases=None):
        self.name = name
        self.aliases = aliases or ()

    @property
    def default_alias(self):
        if len(self.aliases):
            return self.aliases[0]
        return self.name

    @classmethod
    def map(cls, organisms):
        dct = {}
        for organism in organisms:
            dct[organism.name] = organism
            if organism.aliases:
                for alias in organism.aliases:
                    dct[alias] = organism
        return dct

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return isinstance(other, Organism) and self.name == other.name

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, repr(self.name))


class Accession(object):
    def __init__(self, identifier, database=None):
        self.identifier = identifier
        self.database = database

    def __eq__(self, other):
        return isinstance(other, Accession) and \
               self.database == other.database and \
               self.identifier == other.identifier

    def __hash__(self):
        return hash(self.identifier) + hash(self.database)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__,
                               ', '.join('{}={}'.format(key, repr(value))
                                         for key, value in self.__dict__.items() if value))

class Type(object):
    def __init__(self, name, aliases=None):
        self.name = name
        self.aliases = aliases or ()

    @property
    def default_alias(self):
        if len(self.aliases):
            return self.aliases[0]
        return self.name

    @classmethod
    def map(cls, types):
        dct = {}
        for type in types:
            dct[type.name] = type
            if type.aliases:
                for alias in type.aliases:
                    dct[alias] = type
        return dct

    def __eq__(self, other):
        return isinstance(other, Type) and self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, repr(self.name))



from grako.exceptions import GrakoException

from gnomic.models import *
from gnomic.grammar import GnomicParser
from gnomic.semantics import DefaultSemantics
from gnomic.utils import genotype_to_text, genotype_to_string

__all__ = (
    'DEFAULT_TYPES',
    'DEFAULT_ORGANISMS',
    'Genotype',
    'Mutation',
    'FeatureTree',
    'Fusion',
    'Plasmid',
    'Feature',
    'Organism',
    'Accession',
    'Type',
    'encoder'
)

DEFAULT_TYPES = (
    Type('promoter', ('P',)),
    Type('terminator', ('T',)),
    Type('gene')
)

DEFAULT_ORGANISMS = (
    Organism('Escherichia coli', aliases=['Ec', 'E.coli']),
    Organism('Saccaromices cerevisiae', aliases=['Sc', 'S.cerevisiae']),
)


class Genotype(object):
    """


    .. attribute:: raw

        A :class:`tuple` containing the changes introduced in this genotype.

        A change can be one of:

        - A :class:`Mutation`
        - A :class:`Plasmid` (non-integrated plasmid)
        - A :class:`Feature` of type 'phene' with a variant (a "phenotype")

    """

    FUSION_MATCH_WHOLE = 'match-whole-only'
    FUSION_UPDATE_ON_CHANGE = 'update-on-change'
    FUSION_SPLIT_ON_CHANGE = 'split-on-change'
    FUSION_EXPLODE_ON_CHANGE = 'explode-on-change'

    def __init__(self, changes, parent=None, fusion_strategy=FUSION_MATCH_WHOLE):
        self.parent = parent
        self._changes = tuple(changes)

        # TODO FIXME renoval strategy: do not add to removed_features list if there was a match in added_features list
        # TODO same for plasmids!

        sites = set(parent.sites if parent else ())
        markers = set(parent.markers if parent else ())
        phenotypes = set(parent.phenotypes if parent else ())
        added_plasmids = set(parent.added_plasmids if parent else ())
        removed_plasmids = set(parent.removed_plasmids if parent else ())
        added_features = set(parent.added_features if parent else ())
        removed_features = set(parent.removed_features if parent else ())
        added_fusion_features = set(parent.added_fusion_features if parent else ())
        removed_fusion_features = set(parent.removed_fusion_features if parent else ())

        def remove(features, remove, **kwargs):
            return {feature for feature in features if not remove.match(feature, **kwargs)}

        def upsert(features, addition, **kwargs):
            return remove(features, addition, **kwargs) | {addition}

        # removes a feature, but only adds it to removed features if it isn't in added features.
        def remove_or_exclude(added_features, removed_features, exclude):
            for feature in added_features:
                if exclude == feature:
                    return remove(added_features, exclude), removed_features
            return remove(added_features, exclude), upsert(removed_features, exclude)

        for change in changes:
            if isinstance(change, Plasmid):
                # add a plasmid (not integrated)
                added_plasmids |= {change}
                removed_plasmids -= {change}
            elif isinstance(change, Feature):
                # define a certain phenotype or variant of a gene:
                # - rid the genotype of other variants of this feature.
                # - replace the feature within any fusions that contain it or a variant of it.
                phenotypes = upsert(phenotypes, change, match_variant=False)

                added_features = upsert(added_features, change, match_variant=False)
                removed_features = remove(removed_features, change, match_variant=False)

                # fusion-sensitive implementation:
                added_fusion_features = upsert(added_fusion_features, change, match_variant=False)
                removed_fusion_features = remove(removed_fusion_features, change, match_variant=False)
            else:
                # mutation:
                if isinstance(change.old, Plasmid):
                    # deletion of a plasmid
                    added_plasmids, removed_plasmids = remove_or_exclude(added_plasmids,
                                                                         removed_plasmids,
                                                                         change.old)
                elif change.old:
                    # deletion of one (or more) features or fusions
                    for feature in change.old.features():
                        added_features, removed_features = remove_or_exclude(added_features,
                                                                             removed_features,
                                                                             feature)

                    # TODO fusion-sensitive implementation
                    # fusion replace/delete strategies:
                    # FUSION_MATCH_WHOLE_ONLY       | A:B:C D:E  B>X -D:E    A:B:C -B +X
                    # FUSION_UPDATE_ON_CHANGE       | A:B:C B>X              A:X:C     (uses SPLIT on delete)
                    # FUSION_SPLIT_ON_CHANGE        | A:B:C:D C>X            A:B X D
                    # FUSION_EXPLODE_ON_CHANGE      | A:B:C:D C>X            A B X D
                    if fusion_strategy not in (Genotype.FUSION_MATCH_WHOLE,):
                        raise NotImplementedError('Unsupported fusion strategy: {}'.format(fusion_strategy))

                    # TODO: support: fusion split, fusion whole.

                    if fusion_strategy == Genotype.FUSION_MATCH_WHOLE:
                        for feature_or_fusion in change.old:
                            added_fusion_features, removed_fusion_features = remove_or_exclude(added_fusion_features,
                                                                                               removed_fusion_features,
                                                                                               feature_or_fusion)
                    elif fusion_strategy == Genotype.FUSION_SPLIT_ON_CHANGE:
                        pass

                if change.new:
                    # if change.new is a plasmid, the features in the plasmid are integrated
                    if isinstance(change.new, Plasmid):
                        pass  # any further record of the integrated plasmid would go here

                    # insertion of one (or more) features or fusions
                    for feature in change.new.features():
                        removed_features, added_features = remove_or_exclude(removed_features,
                                                                             added_features,
                                                                             feature)

                        # added_features = upsert(added_features, feature)
                        # removed_features = remove(removed_features, feature)

                    # fusion-sensitive implementation:
                    for feature_or_fusion in change.new:
                        added_fusion_features = upsert(added_fusion_features, feature_or_fusion)
                        removed_fusion_features = remove(removed_fusion_features, feature_or_fusion)

                if change.old and change.new:
                    # in a replacement, the removed part must be a single feature
                    sites = upsert(sites, change.old.contents[0])

                if change.markers:
                    for marker in change.markers:
                        markers = upsert(markers, marker, match_variant=False)
                        added_features = upsert(added_features, marker, match_variant=False)
                        removed_features = remove(removed_features, marker, match_variant=False)
                        added_fusion_features = upsert(added_fusion_features, marker, match_variant=False)
                        removed_fusion_features = remove(removed_fusion_features, marker, match_variant=False)

        self.sites = tuple(sites)
        self.markers = tuple(markers)
        self.phenotypes = tuple(phenotypes)
        self.added_plasmids = tuple(added_plasmids)
        self.removed_plasmids = tuple(removed_plasmids)
        self.added_features = tuple(added_features)
        self.removed_features = tuple(removed_features)
        self.added_fusion_features = tuple(added_fusion_features)
        self.removed_fusion_features = tuple(removed_fusion_features)

    @property
    def raw(self):
        return self._changes

    @classmethod
    def _parse_string(cls, string, *args, **kwargs):
        parser = GnomicParser()
        semantics = DefaultSemantics(*args, **kwargs)
        return parser.parse(string,
                            whitespace='',
                            semantics=semantics,
                            rule_name='start')

    @classmethod
    def is_valid(cls, string):
        """
        Tests whether a gnomic genotype definition can be parsed.

        :param str string: The genotype definition to validate.
        :return: :class:`bool`
        """
        try:
            cls._parse_string(string)
            return True
        except GrakoException:
            return False

    @classmethod
    def parse(cls,
              string,
              parent=None,
              organisms=None,
              types=None,
              **kwargs):
        """

        :param str string: The genotype definition to parse
        :param Genotype parent:
        :param organisms: A list or tuple of :class:`Organism` objects used to resolve organism designations in
            features. If empty, uses: :attr:`DEFAULT_ORGANISMS`.
        :param types: A list or tuple of :class:`Type` objects used to resolve type designations in features.
            If empty, uses: :attr:`DEFAULT_TYPES`.
        :param kwargs: Any additional attributes, such as `fusion_strategy`, used with :class:`Genotype`.
        :return: A :class:`Genotype` instance with the parsed gnomic genotype.
        """
        changes = cls._parse_string(string,
                                    organisms or DEFAULT_ORGANISMS,
                                    types or DEFAULT_TYPES)
        return Genotype(changes, parent=parent, **kwargs)

    @staticmethod
    def chain_parse(definitions, **kwargs):
        genotype = Genotype.parse(definitions[0], **kwargs)
        for definition in definitions[1:]:
            genotype = Genotype.parse(definition, parent=genotype, **kwargs)
        return genotype

    def _iter_changes(self, fusions=True):
        if fusions:
            for feature in self.added_fusion_features:
                yield Ins(feature)

            for feature in self.removed_fusion_features:
                yield Del(feature)
        else:
            for feature in self.added_features:
                yield Ins(feature)

            for feature in self.removed_features:
                yield Del(feature)

        for plasmid in self.added_plasmids:
            yield plasmid

        for plasmid in self.removed_plasmids:
            yield Del(plasmid)

    # TODO some getter for accessing the original changes (self._changes)

    def changes(self, fusions=False):
        """
        :param bool fusions: Keeps fusions together if ``True``, otherwise includes them as individual features.
        :return:
        """
        return set(self._iter_changes(fusions=fusions))

    def format(self, fusions=True, output='text'):
        """
        :param bool fusions: Keeps fusions together if ``True``, otherwise includes them as individual features.
        :param str output: Output format; one of ``'text'`` and ``'string'``
        :return:
        """
        if output == 'text':
            return genotype_to_text(self, fusions=fusions)
        elif output == 'string':
            return genotype_to_string(self, fusions=fusions)
        else:
            raise NotImplementedError('Unknown output format: {}'.format(output))

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, repr({'parent': self.parent, 'changes': self._changes}))


import json

from gnomic.models import Fusion, Organism, Accession


class GnomicJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Fusion):
            return {
                "fusion": o.contents
            }
        elif isinstance(o, Organism):
            return o.default_alias
        elif isinstance(o, Accession):
            return {"db": o.database, "id": o.identifier}
        # TODO full JSON encoding
        try:
            iterable = iter(o)
        except TypeError:
            pass
        else:
            return list(iterable)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, o)


from gnomic.models import Mutation, Fusion, Plasmid, Feature, Organism, Accession, Type, FeatureTree, Range, FeatureSet
from gnomic.grammar import GnomicSemantics


class DefaultSemantics(GnomicSemantics):
    def __init__(self,
                 organisms=None,
                 types=None):
        self._organisms = {} if organisms is None else Organism.map(organisms)
        self._types = {} if types is None else Type.map(types)

    def FUSION(self, ast):
        return Fusion(*ast)

    def FEATURE_FUSION(self, ast):
        return Fusion(*ast)

    def FEATURE_SET(self, ast):
        return FeatureSet(*ast)

    def ORGANISM(self, name):
        try:
            return self._organisms[name]
        except KeyError:
            self._organisms[name] = organism = Organism(name)
            return organism

    def BINARY_VARIANT(self, variant):
        if variant == '+':
            return 'wild-type'
        else:
            return 'mutant'

    def DNA_SEQUENCE_VARIANT(self, ast):
        return ''.join(map(str, ast))

    def PROTEIN_SEQUENCE_VARIANT(self, ast):
        return ''.join(map(str, ast))

    def SEQUENCE_VARIANT(self, ast):
        return ''.join(ast)

    def insertion(self, ast):
        return Mutation(None, ast.new, markers=ast.markers)

    def replacement(self, ast):
        return Mutation(ast.old,
                        ast.new,
                        markers=ast.markers,
                        multiple=ast.op == '>>')

    def deletion(self, ast):
        return Mutation(ast.old, None, markers=ast.markers)

    def RANGE(self, ast):
        level = {
            'c': 'coding',
            'r': 'RNA',
            'p': 'protein'
        }[ast.level]

        if ast.pos:
            return Range(level, ast.pos, ast.pos)
        return Range(level, ast.start, ast.end)

    def INTEGER(self, ast):
        return int(ast)

    def ACCESSION(self, ast):
        return Accession(ast['id'], ast['db'])

    def PLASMID(self, ast):
        return Plasmid(ast.name, ast.contents, markers=ast.markers)

    def PHENE(self, ast):
        return self.FEATURE(ast, default_type='phene')

    def FEATURE(self, ast, default_type=None):
        if ast.type or default_type:
            name = ast.type or default_type
            try:
                type = self._types[name]
            except KeyError:
                self._types[name] = type = Type(name)
        else:
            type = None

        return Feature(ast.name, type,
                       accession=ast.accession,
                       organism=ast.organism,
                       variant=', '.join(ast.variant) if isinstance(ast.variant, list) else ast.variant,
                       range=ast.range)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# CAVEAT UTILITOR
#
# This file was automatically generated by Grako.
#
#    https://pypi.python.org/pypi/grako/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.


from __future__ import print_function, division, absolute_import, unicode_literals

from grako.buffering import Buffer
from grako.parsing import graken, Parser
from grako.util import re, RE_FLAGS, generic_main  # noqa


__all__ = [
    'GnomicParser',
    'GnomicSemantics',
    'main'
]

KEYWORDS = {}


class GnomicBuffer(Buffer):
    def __init__(
        self,
        text,
        whitespace=re.compile('[\\t ]+', RE_FLAGS | re.DOTALL),
        nameguard=None,
        comments_re=None,
        eol_comments_re=None,
        ignorecase=None,
        namechars='',
        **kwargs
    ):
        super(GnomicBuffer, self).__init__(
            text,
            whitespace=whitespace,
            nameguard=nameguard,
            comments_re=comments_re,
            eol_comments_re=eol_comments_re,
            ignorecase=ignorecase,
            namechars=namechars,
            **kwargs
        )


class GnomicParser(Parser):
    def __init__(
        self,
        whitespace=re.compile('[\\t ]+', RE_FLAGS | re.DOTALL),
        nameguard=None,
        comments_re=None,
        eol_comments_re=None,
        ignorecase=None,
        left_recursion=False,
        parseinfo=True,
        keywords=None,
        namechars='',
        buffer_class=GnomicBuffer,
        **kwargs
    ):
        if keywords is None:
            keywords = KEYWORDS
        super(GnomicParser, self).__init__(
            whitespace=whitespace,
            nameguard=nameguard,
            comments_re=comments_re,
            eol_comments_re=eol_comments_re,
            ignorecase=ignorecase,
            left_recursion=left_recursion,
            parseinfo=parseinfo,
            keywords=keywords,
            namechars=namechars,
            buffer_class=buffer_class,
            **kwargs
        )

    @graken()
    def _SEQUENCE_VARIANT_(self):
        with self._choice():
            with self._option():
                self._DNA_SEQUENCE_VARIANT_()
            with self._option():
                self._PROTEIN_SEQUENCE_VARIANT_()
            self._error('no available options')

    @graken()
    def _DNA_SEQUENCE_VARIANT_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._token('g')
                with self._option():
                    self._token('c')
                with self._option():
                    self._token('n')
                self._error('expecting one of: c g n')
        self._token('.')
        with self._group():
            with self._choice():
                with self._option():
                    self._INTEGER_()
                    self._NUCLEOTIDE_()
                    self._token('>')
                    self._NUCLEOTIDE_()
                with self._option():
                    self._INTEGER_()
                    self._token('_')
                    self._INTEGER_()
                    self._token('del')
                with self._option():
                    self._INTEGER_()
                    self._token('_')
                    self._INTEGER_()
                    self._token('ins')
                    self._NUCLEOTIDE_SEQUENCE_()
                with self._option():
                    self._INTEGER_()
                    self._token('_')
                    self._INTEGER_()
                    self._token('delins')
                    self._NUCLEOTIDE_SEQUENCE_()
                with self._option():
                    self._INTEGER_()
                    self._token('_')
                    self._INTEGER_()
                    self._token('dup')
                self._error('no available options')

    @graken()
    def _NUCLEOTIDE_SEQUENCE_(self):
        self._NUCLEOTIDE_()

        def block0():
            self._NUCLEOTIDE_()
        self._closure(block0)

    @graken()
    def _NUCLEOTIDE_(self):
        self._pattern(r'[ACGTBDHKMNRSVWY]')

    @graken()
    def _PROTEIN_SEQUENCE_VARIANT_(self):
        self._token('p.')
        with self._group():
            with self._choice():
                with self._option():
                    self._AMINO_ACID_()
                    self._INTEGER_()
                    self._AMINO_ACID_()
                with self._option():
                    self._AMINO_ACID_()
                    self._INTEGER_()
                    self._token('*')
                self._error('no available options')

    @graken()
    def _AMINO_ACID_SEQUENCE_(self):
        self._AMINO_ACID_()

        def block0():
            self._AMINO_ACID_()
        self._closure(block0)

    @graken()
    def _AMINO_ACID_(self):
        self._pattern(r'[A-Z]([a-z]{2})?')

    @graken()
    def _start_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._change_()
                    self.add_last_node_to_name('@')

                    def block1():
                        with self._group():
                            with self._choice():
                                with self._option():
                                    self._sep_()
                                with self._option():
                                    self._list_separator_()
                                    with self._optional():
                                        self._sep_()
                                self._error('no available options')
                        self._change_()
                        self.add_last_node_to_name('@')
                    self._closure(block1)
                with self._option():
                    self._empty_closure()
                self._error('no available options')
        with self._optional():
            self._sep_()
        self._check_eof()

    @graken()
    def _change_(self):
        with self._choice():
            with self._option():
                self._insertion_()
            with self._option():
                self._replacement_()
            with self._option():
                self._deletion_()
            with self._option():
                self._PLASMID_()
            with self._option():
                self._PHENE_()
            self._error('no available options')

    @graken()
    def _insertion_(self):
        self._token('+')
        self._INSERTABLE_()
        self.name_last_node('new')
        with self._optional():
            self._MARKERS_()
            self.name_last_node('markers')
        self.ast._define(
            ['markers', 'new'],
            []
        )

    @graken()
    def _replacement_(self):
        with self._choice():
            with self._option():
                self._REPLACEABLE_()
                self.name_last_node('old')
                self._token('>')
                self.name_last_node('op')
                self._SUBSTITUTE_()
                self.name_last_node('new')
                with self._optional():
                    self._MARKERS_()
                    self.name_last_node('markers')
            with self._option():
                self._REPLACEABLE_()
                self.name_last_node('old')
                self._token('>>')
                self.name_last_node('op')
                self._SUBSTITUTE_()
                self.name_last_node('new')
                with self._optional():
                    self._MARKERS_()
                    self.name_last_node('markers')
            self._error('no available options')
        self.ast._define(
            ['markers', 'new', 'old', 'op'],
            []
        )

    @graken()
    def _deletion_(self):
        self._token('-')
        self._DELETABLE_()
        self.name_last_node('old')
        with self._optional():
            self._MARKERS_()
            self.name_last_node('markers')
        self.ast._define(
            ['markers', 'old'],
            []
        )

    @graken()
    def _INSERTABLE_(self):
        with self._choice():
            with self._option():
                self._FUSION_()
            with self._option():
                self._FEATURE_SET_()
                self.name_last_node('@')
            with self._option():
                self._FEATURE_()
            self._error('no available options')

    @graken()
    def _REPLACEABLE_(self):
        with self._choice():
            with self._option():
                self._FUSION_()
            with self._option():
                self._FEATURE_SET_()
                self.name_last_node('@')
            with self._option():
                self._FEATURE_()
            self._error('no available options')

    @graken()
    def _SUBSTITUTE_(self):
        with self._choice():
            with self._option():
                self._PLASMID_()
            with self._option():
                self._FUSION_()
            with self._option():
                self._FEATURE_SET_()
                self.name_last_node('@')
            with self._option():
                self._FEATURE_()
            self._error('no available options')

    @graken()
    def _DELETABLE_(self):
        with self._choice():
            with self._option():
                self._PLASMID_()
            with self._option():
                self._FUSION_()
            with self._option():
                self._FEATURE_SET_()
                self.name_last_node('@')
            with self._option():
                self._FEATURE_()
            self._error('no available options')

    @graken()
    def _PLASMID_(self):
        with self._choice():
            with self._option():
                self._IDENTIFIER_()
                self.name_last_node('name')
                self._FEATURE_SET_()
                self.name_last_node('contents')
                with self._optional():
                    self._MARKERS_()
                    self.name_last_node('markers')
            with self._option():
                self._IDENTIFIER_()
                self.name_last_node('name')
                self._token('{}')
                with self._optional():
                    self._MARKERS_()
                    self.name_last_node('markers')
            with self._option():
                self._token('(')
                self._IDENTIFIER_()
                self.name_last_node('name')
                self._FEATURE_SET_ITEMS_()
                self.name_last_node('contents')
                self._token(')')
                with self._optional():
                    self._MARKERS_()
                    self.name_last_node('markers')
            with self._option():
                self._token('(')
                self._IDENTIFIER_()
                self.name_last_node('name')
                self._token(')')
                with self._optional():
                    self._MARKERS_()
                    self.name_last_node('markers')
            self._error('no available options')
        self.ast._define(
            ['contents', 'markers', 'name'],
            []
        )

    @graken()
    def _PHENE_(self):
        with self._choice():
            with self._option():
                with self._optional():
                    self._FEATURE_ORGANISM_()
                    self.name_last_node('organism')
                self._IDENTIFIER_()
                self.name_last_node('name')
                self._ACCESSION_()
                self.name_last_node('accession')
                self._BINARY_VARIANT_()
                self.name_last_node('variant')
            with self._option():
                self._ACCESSION_()
                self.name_last_node('accession')
                self._BINARY_VARIANT_()
                self.name_last_node('variant')
            with self._option():
                with self._optional():
                    self._FEATURE_ORGANISM_()
                    self.name_last_node('organism')
                self._IDENTIFIER_()
                self.name_last_node('name')
                self._VARIANT_()
                self.name_last_node('variant')
            self._error('no available options')
        self.ast._define(
            ['accession', 'name', 'organism', 'variant'],
            []
        )

    @graken()
    def _PHENE_SET_(self):
        self._token('{')
        with self._optional():
            self._sep_()
        self._PHENE_()
        self.add_last_node_to_name('@')

        def block1():
            with self._group():
                with self._choice():
                    with self._option():
                        self._sep_()
                    with self._option():
                        self._list_separator_()
                        with self._optional():
                            self._sep_()
                    self._error('no available options')
            self._PHENE_()
            self.add_last_node_to_name('@')
        self._closure(block1)
        with self._optional():
            self._sep_()
        self._token('}')

    @graken()
    def _MARKERS_(self):
        self._token('::')
        with self._group():
            with self._choice():
                with self._option():
                    self._PHENE_()
                    self.add_last_node_to_name('@')
                with self._option():
                    self._PHENE_SET_()
                    self.name_last_node('@')
                self._error('no available options')

    @graken()
    def _FEATURE_(self):
        with self._choice():
            with self._option():
                with self._optional():
                    self._FEATURE_ORGANISM_()
                    self.name_last_node('organism')
                with self._optional():
                    self._IDENTIFIER_()
                    self.name_last_node('type')
                    self._token('.')
                self._IDENTIFIER_()
                self.name_last_node('name')
                with self._optional():
                    self._VARIANT_()
                    self.name_last_node('variant')
                with self._optional():
                    self._ACCESSION_()
                    self.name_last_node('accession')
                with self._optional():
                    self._RANGE_()
                    self.name_last_node('range')
            with self._option():
                self._ACCESSION_()
                self.name_last_node('accession')
                with self._optional():
                    self._RANGE_()
                    self.name_last_node('range')
            self._error('no available options')
        self.ast._define(
            ['accession', 'name', 'organism', 'range', 'type', 'variant'],
            []
        )

    @graken()
    def _FEATURE_ORGANISM_(self):
        self._ORGANISM_()
        self.name_last_node('@')
        self._token('/')

    @graken()
    def _FEATURE_SET_(self):
        self._token('{')
        self._FEATURE_SET_ITEMS_()
        self.name_last_node('@')
        self._token('}')

    @graken()
    def _FEATURE_SET_ITEMS_(self):
        with self._optional():
            self._sep_()
        with self._group():
            with self._choice():
                with self._option():
                    self._FEATURE_FUSION_()
                with self._option():
                    self._FEATURE_()
                self._error('no available options')
        self.add_last_node_to_name('@')

        def block2():
            with self._group():
                with self._choice():
                    with self._option():
                        self._sep_()
                    with self._option():
                        self._list_separator_()
                        with self._optional():
                            self._sep_()
                    self._error('no available options')
            with self._group():
                with self._choice():
                    with self._option():
                        self._FEATURE_FUSION_()
                    with self._option():
                        self._FEATURE_()
                    self._error('no available options')
            self.add_last_node_to_name('@')
        self._closure(block2)
        with self._optional():
            self._sep_()

    @graken()
    def _FUSION_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._FEATURE_SET_()
                with self._option():
                    self._FEATURE_()
                self._error('no available options')
        self.add_last_node_to_name('@')

        def block2():
            self._token(':')
            with self._group():
                with self._choice():
                    with self._option():
                        self._FEATURE_SET_()
                    with self._option():
                        self._FEATURE_()
                    self._error('no available options')
            self.add_last_node_to_name('@')
        self._positive_closure(block2)

    @graken()
    def _FEATURE_FUSION_(self):
        self._FEATURE_()
        self.add_last_node_to_name('@')

        def block1():
            self._token(':')
            self._FEATURE_()
            self.add_last_node_to_name('@')
        self._positive_closure(block1)

    @graken()
    def _VARIANT_(self):
        with self._choice():
            with self._option():
                self._token('(')
                self._VARIANT_LIST_()
                self.name_last_node('@')
                self._token(')')
            with self._option():
                self._BINARY_VARIANT_()
                self.name_last_node('@')
            self._error('no available options')

    @graken()
    def _VARIANT_LIST_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._SEQUENCE_VARIANT_()
                with self._option():
                    self._VARIANT_IDENTIFIER_()
                self._error('no available options')
        self.name_last_node('@')

        def block2():
            with self._group():
                with self._choice():
                    with self._option():
                        self._token(',')
                    with self._option():
                        self._token(';')
                    self._error('expecting one of: , ;')
            with self._optional():
                self._sep_()
            with self._group():
                with self._choice():
                    with self._option():
                        self._SEQUENCE_VARIANT_()
                    with self._option():
                        self._VARIANT_IDENTIFIER_()
                    self._error('no available options')
            self.name_last_node('@')
        self._closure(block2)

    @graken()
    def _BINARY_VARIANT_(self):
        with self._choice():
            with self._option():
                self._token('+')
            with self._option():
                self._token('-')
            self._error('expecting one of: + -')

    @graken()
    def _VARIANT_IDENTIFIER_(self):
        self._pattern(r'[A-Za-z0-9]+([A-Za-z0-9_\-]+[A-Za-z0-9])?')

    @graken()
    def _RANGE_(self):
        with self._choice():
            with self._option():
                self._token('[')
                with self._optional():
                    self._SEQUENCE_LEVEL_PREFIX_()
                    self.name_last_node('level')
                self._INTEGER_()
                self.name_last_node('start')
                self._token('_')
                self._INTEGER_()
                self.name_last_node('end')
                self._token(']')
            with self._option():
                self._token('[')
                with self._optional():
                    self._SEQUENCE_LEVEL_PREFIX_()
                    self.name_last_node('level')
                self._INTEGER_()
                self.name_last_node('pos')
                self._token(']')
            self._error('no available options')
        self.ast._define(
            ['end', 'level', 'pos', 'start'],
            []
        )

    @graken()
    def _SEQUENCE_LEVEL_PREFIX_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._token('g')
                with self._option():
                    self._token('c')
                with self._option():
                    self._token('p')
                self._error('expecting one of: c g p')
        self.name_last_node('@')
        self._token('.')

    @graken()
    def _ACCESSION_(self):
        with self._choice():
            with self._option():
                self._token('#')
                self._DATABASE_()
                self.name_last_node('db')
                self._token(':')
                with self._group():
                    with self._choice():
                        with self._option():
                            self._INTEGER_()
                        with self._option():
                            self._IDENTIFIER_()
                        self._error('no available options')
                self.name_last_node('id')
            with self._option():
                self._token('#')
                with self._group():
                    with self._choice():
                        with self._option():
                            self._INTEGER_()
                        with self._option():
                            self._IDENTIFIER_()
                        self._error('no available options')
                self.name_last_node('id')
            self._error('no available options')
        self.ast._define(
            ['db', 'id'],
            []
        )

    @graken()
    def _DATABASE_(self):
        self._pattern(r'[A-Za-z0-9-][A-Za-z0-9]+')

    @graken()
    def _INTEGER_(self):
        self._pattern(r'[0-9]+')
        self.name_last_node('@')

    @graken()
    def _IDENTIFIER_(self):
        self._pattern(r'[A-Za-z0-9]+([A-Za-z0-9_-]+[A-Za-z0-9])?')

    @graken()
    def _ORGANISM_(self):
        self._pattern(r'[a-zA-Z0-9]+(\.[a-zA-Z0-9]+)?')

    @graken()
    def _list_separator_(self):
        self._token(',')

    @graken()
    def _sep_(self):
        self._pattern(r'[\t ]+')


class GnomicSemantics(object):
    def SEQUENCE_VARIANT(self, ast):
        return ast

    def DNA_SEQUENCE_VARIANT(self, ast):
        return ast

    def NUCLEOTIDE_SEQUENCE(self, ast):
        return ast

    def NUCLEOTIDE(self, ast):
        return ast

    def PROTEIN_SEQUENCE_VARIANT(self, ast):
        return ast

    def AMINO_ACID_SEQUENCE(self, ast):
        return ast

    def AMINO_ACID(self, ast):
        return ast

    def start(self, ast):
        return ast

    def change(self, ast):
        return ast

    def insertion(self, ast):
        return ast

    def replacement(self, ast):
        return ast

    def deletion(self, ast):
        return ast

    def INSERTABLE(self, ast):
        return ast

    def REPLACEABLE(self, ast):
        return ast

    def SUBSTITUTE(self, ast):
        return ast

    def DELETABLE(self, ast):
        return ast

    def PLASMID(self, ast):
        return ast

    def PHENE(self, ast):
        return ast

    def PHENE_SET(self, ast):
        return ast

    def MARKERS(self, ast):
        return ast

    def FEATURE(self, ast):
        return ast

    def FEATURE_ORGANISM(self, ast):
        return ast

    def FEATURE_SET(self, ast):
        return ast

    def FEATURE_SET_ITEMS(self, ast):
        return ast

    def FUSION(self, ast):
        return ast

    def FEATURE_FUSION(self, ast):
        return ast

    def VARIANT(self, ast):
        return ast

    def VARIANT_LIST(self, ast):
        return ast

    def BINARY_VARIANT(self, ast):
        return ast

    def VARIANT_IDENTIFIER(self, ast):
        return ast

    def RANGE(self, ast):
        return ast

    def SEQUENCE_LEVEL_PREFIX(self, ast):
        return ast

    def ACCESSION(self, ast):
        return ast

    def DATABASE(self, ast):
        return ast

    def INTEGER(self, ast):
        return ast

    def IDENTIFIER(self, ast):
        return ast

    def ORGANISM(self, ast):
        return ast

    def list_separator(self, ast):
        return ast

    def sep(self, ast):
        return ast


def main(filename, startrule, **kwargs):
    with open(filename) as f:
        text = f.read()
    parser = GnomicParser(parseinfo=False)
    return parser.parse(text, startrule, filename=filename, **kwargs)


if __name__ == '__main__':
    import json
    ast = generic_main(main, GnomicParser, name='Gnomic')
    print('AST:')
    print(ast)
    print()
    print('JSON:')
    print(json.dumps(ast, indent=2))
    print()



import collections
from gnomic.models import Mutation, Plasmid, Fusion, FeatureTree, FeatureSet, Feature


def namedtuple_with_defaults(typename, field_names, default_values=()):
    T = collections.namedtuple(typename, field_names)
    T.__new__.__defaults__ = (None,) * len(T._fields)
    if isinstance(default_values, collections.Mapping):
        prototype = T(**default_values)
    else:
        prototype = T(*default_values)
    T.__new__.__defaults__ = tuple(prototype)
    return T


def genotype_to_text(genotype, fusions=False):
    """
    A method to create a more biologist friendly representation
    of the genotype string
    :param genotype: An instance of the Genotype class
    :param bool fusions: Whether to keep fusions together.
    :param delta_char: This symbol is used to display a deletion
    :return: str
    """
    parts = []
    for change in genotype.changes(fusions=fusions):
        parts.append(change_to_text(change))

    return " ".join(parts)


def change_to_text(change, delta_char=u"\u0394"):
    s = None
    change_type = type(change)
    if change_type is Mutation:
        if change.old and change.new:  # substitution
            s = delta_char + feature_to_text(change.old) + '::' + feature_to_text(change.new)
        elif change.old is None:  # insertion
            s = feature_to_text(change.new)
        elif change.new is None:  # deletion
            s = delta_char + feature_to_text(change.old, integrated=False)
    elif change_type is Plasmid:
        s = feature_to_text(change, integrated=False)

    if change.markers:
        s += "::" + feature_to_text(change, is_maker=True)

    return s


def feature_to_text(feature, integrated=True, is_maker=False):
    """
    A method to transform a genotype feature into text
    :param feature: Genotype feature
    :param integrated: boolean
    :param is_maker: boolean
    :return: str
    """
    feature_type = type(feature)
    if feature_type is Plasmid:
        name = feature.name
        if feature.contents:
            content = ' '.join(map(feature_to_text, feature.contents))

            if integrated:
                return '%s(%s)' % (name, content)
            else:
                return '(%s %s)' % (name, content)

        elif integrated:
            return name
        else:
            return '(%s)' % name

    elif feature_type is Fusion:
        return ':'.join(map(feature_to_text, feature.contents))

    elif feature_type is FeatureTree:
        return ' '.join(map(feature_to_text, feature.contents))

    else:
        text = ''
        if is_maker:
            text += '::'

        if feature.organism:
            text += '%s/' % feature.organism.name

        text += feature.name

        variant_map = {'wild-type': u"\u207A",
                       'mutant': u"\u207B"}
        variant = feature.variant

        if variant:
            if variant in variant_map:
                text += variant_map[variant]
            else:
                text += "(%s)" % variant

        return text


def genotype_to_string(genotype, fusions=True):
    parts = []

    for change in genotype.changes(fusions=fusions):
        parts.append(change_to_string(change))
    return ' '.join(parts)


def change_to_string(change):
    s = None

    if isinstance(change, Mutation):
        if change.old and change.new:
            if change.multiple:
                s = '{}>>{}'.format(feature_to_string(change.old), feature_to_string(change.new))
            else:
                s = '{}>{}'.format(feature_to_string(change.old), feature_to_string(change.new))
        elif change.old is None:
            # FIXME phenotypes should not have a + or - prefix; only a postfix.
            s = '+{}'.format(feature_to_string(change.new))
        elif change.new is None:
            s = '-{}'.format(feature_to_string(change.old))
    elif isinstance(change, Plasmid):
        s = feature_to_string(change)
    else:
        raise TypeError()

    if change.markers:
        s += '::{}'.format(feature_to_string(change.markers))

    return s


def feature_to_string(feature):
    if isinstance(feature, Plasmid):
        name = feature.name
        if feature.contents:
            contents = ' '.join(feature_to_string(f) for f in feature.contents)
            return r'({name} {contents})'.format(name=name, contents=contents)
        else:
            return '({name})'.format(name=name)
    elif isinstance(feature, Fusion):
        return ':'.join(feature_to_string(f) for f in feature.contents)
    elif isinstance(feature, FeatureTree):
        contents = ' '.join(feature_to_string(f) for f in feature.contents)
        if isinstance(feature, FeatureSet):
            return '{{{contents}}}'.format(contents=contents)
        elif len(feature) != 1:
            return '{{{contents}}}'.format(contents=contents)
        else:
            return contents
    else:
        s = ''
        if feature.type and feature.type.name != 'phene':
            s += '{}.'.format(feature.type.name)

        if feature.organism:
            s += '{}/'.format(feature.organism.name)

        s += feature.name

        variant = feature.variant
        variant_map = {'wild-type': '+', 'mutant': '-'}

        if variant:
            try:
                s += variant_map[variant]
            except KeyError:
                s += '({})'.format(variant)
        return s



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import codecs
from setuptools import setup, find_packages

setup(
    name='gnomic',
    version='0.4.0',
    packages=find_packages(exclude=['*tests*']),
    license='Apache',
    author='Lars Schöning',
    author_email='lays@biosustain.dtu.dk',
    description='A grammar for describing microbial genotypes and phenotypes',
    long_description=codecs.open('README.rst', encoding='utf-8').read(),
    test_suite='nose.collector',
    tests_require=['nose'],
    install_requires=[
        'grako>=3.18.1',
        'six>=1.8.0'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
    extras_require={
        'docs': ['Sphinx', 'sphinx-rtd-theme'],
    }
)
